from abc import ABC, abstractmethod

# //////////////////////////////////////////////////////////////////////////////
class Backend(ABC):
    """Abstract base class for terminal backends.
    This class defines the interface for terminal operations, such as writing text,
    handling colors, and managing terminal size and key input.
    Subclasses must implement these methods to provide specific terminal functionality."""

    def __init__(self):
        self._nodelay_mode: bool = False

    @abstractmethod
    def set_nodelay(self, boolean: bool) -> None:
        """Set the nodelay mode for the backend. When enabled, get_key() will not block the terminal."""
        raise NotImplementedError

    @abstractmethod
    def sleep(self, ms: int) -> None:
        """Sleep for a given number of milliseconds."""
        raise NotImplementedError

    @abstractmethod
    def write_text(self, y: int, x: int, chars: str, attr: int = 0) -> None:
        """Write text to the terminal at a specific position."""
        raise NotImplementedError

    @abstractmethod
    def get_size(self, update = False) -> tuple[int,int]:
        """Get the size of the terminal, as (COLS, LINES).
        If update is True, a backend method will be called to update the current COLS and LINES values."""
        raise NotImplementedError

    @abstractmethod
    def supports_color(self) -> bool:
        """Check if the backend supports color."""
        raise NotImplementedError

    @abstractmethod
    def init_color(self, i: int, r: int, g: int, b: int) -> None:
        """Initialize a color with index 'i' and values 'r','g','b' to be used in color pairs by the terminal."""
        raise NotImplementedError

    @abstractmethod
    def init_pair(self, i: int, fg: int, bg: int) -> None:
        """Initialize a color pair (fg,bg) which can be accessed with index 'i'."""
        raise NotImplementedError

    @abstractmethod
    def get_color_pair(self, i: int) -> int:
        """Retrieve the color pair for a given index."""
        raise NotImplementedError

    @abstractmethod
    def _start(self) -> None:
        """Initialize the backend, setting up the terminal."""
        raise NotImplementedError

    @abstractmethod
    def _end(self) -> None:
        """Clean up the backend, restoring the terminal to its original state."""
        raise NotImplementedError

    @abstractmethod
    def _refresh(self) -> None:
        """Refresh the terminal display."""
        raise NotImplementedError

    @abstractmethod
    def _get_key(self) -> int:
        """Get a key press from the terminal."""
        raise NotImplementedError

    @abstractmethod
    def _resize(self, h: int, w: int) -> None:
        """Resize the terminal to the given height and width."""
        raise NotImplementedError


# //////////////////////////////////////////////////////////////////////////////



import json

import prismatui as pr

# //////////////////////////////////////////////////////////////////////////////
class Palette:
    """Class to manage colors and palette files."""
    def __init__(self):
        self.palette: dict = {"colors": [], "pairs":  []}


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def set_colors(self, colors) -> None:
        """Update the color palette with a list of (r,g,b) colors.
        RBG values should be in the range 0-1000."""
        self.palette["colors"] = [[int(c) for c in color] for color in colors]

    # --------------------------------------------------------------------------
    def set_pairs(self, pairs) -> None:
        """Update the color pairs with a list of (fg, bg) pairs."""
        self.palette["pairs"] = [[int(p) for p in pair] for pair in pairs]


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def save_pal(self, path_pal: str) -> None:
        """Save the current palette to a JSON file."""
        with open(path_pal, 'w') as file:
            json.dump(self.palette, file)

    # --------------------------------------------------------------------------
    def load_pal(self, path_pal: str) -> None:
        """Load a palette from a JSON file."""
        with open(path_pal, 'r') as file:
            palette_dict = json.load(file)
        self.load_dict(palette_dict)

    # --------------------------------------------------------------------------
    def load_dict(self, palette_dict: dict[str, list[tuple]]) -> None:
        """
        Load a palette from a dictionary. It must contain 'colors' and 'pairs' keys.
        Colors should be a list of (r,g,b) values in the range 0-1000.
        Pairs should be a list of (fg,bg) indices.
        """
        self.palette = palette_dict.copy()
        colors = self.palette["colors"]
        pairs  = self.palette["pairs"]

        assert len(colors) <= pr.MAX_PALETTE_COLORS, \
            f"Palette has {len(colors)} colors, max is {pr.MAX_PALETTE_COLORS}."

        assert len(pairs) <= pr.MAX_PALETTE_PAIRS, \
            f"Palette has {len(pairs)} pairs, max is {pr.MAX_PALETTE_PAIRS}."

        if not pr._CURRENT_BACKEND.supports_color(): return

        for i,color in enumerate(colors):
            pr._CURRENT_BACKEND.init_color(i, *color)

        for i,(fg,bg) in enumerate(pairs):
            pr._CURRENT_BACKEND.init_pair(i, fg, bg)


# //////////////////////////////////////////////////////////////////////////////



from typing import Generator

import prismatui as pr

# //////////////////////////////////////////////////////////////////////////////
class Layer:
    """Layer class for managing pixel data in a 2D grid."""
    def __init__(self,
        h: int, w: int,
        chars: list[str] = None,
        attrs: list[list[int]] = None,
        blend = pr.BlendMode.OVERLAY
    ):
        if chars is None: chars = ((pr.BLANK_CHAR for _ in range(w)) for _ in range(h))
        if attrs is None: attrs = ((pr.BLANK_ATTR for _ in range(w)) for _ in range(h))
        self.h: int = h
        self.w: int = w
        self._data: list[list[pr.Pixel]] = self.get_pixel_mat(chars, attrs)
        self.blend_mode: pr.BlendMode = blend


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    @classmethod
    def get_pixel_mat(cls, chars: list[str], attrs: list[list[int]]) -> list[list["pr.Pixel"]]:
        """Create a pixel matrix from a matrix of characters and a matrix of attributes."""
        return [
            [pr.Pixel(c,a) for c,a in zip(row_chars, row_attrs)]
            for row_chars, row_attrs in zip(chars, attrs)
        ]


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def set_size(self, h: int, w: int) -> None:
        """Set the size of the layer, adjusting the pixel matrix accordingly."""
        if   h < self.h: self._remove_rows(h)
        elif h > self.h: self._add_rows(h - self.h)
        self.h = h

        if   w < self.w: self._remove_cols(w)
        elif w > self.w: self._add_cols(w - self.w)
        self.w = w

    # --------------------------------------------------------------------------
    def get_chars(self) -> list[list[str]]:
        """Get the characters of the layer as a 2D list of chars."""
        return [[pixel._char for pixel in row] for row in self._data]

    # --------------------------------------------------------------------------
    def get_chars_row_as_strs(self) -> list[str]:
        """Get the characters of the layer as a list of strings, one for each row."""
        return [''.join(pixel._char for pixel in row) for row in self._data]

    # --------------------------------------------------------------------------
    def get_attrs(self) -> list[list[int]]:
        """Get the attributes of the layer as a 2D list of ints."""
        return [[pixel._attr for pixel in row] for row in self._data]

    # --------------------------------------------------------------------------
    def copy(self) -> "Layer":
        """Create a copy of the layer."""
        return Layer(self.h, self.w, self.get_chars(), self.get_attrs(), self.blend_mode)

    # --------------------------------------------------------------------------
    def clear(self) -> None:
        """Clear the layer, resetting all pixels to blank."""
        self._data = [self._create_row(self.w) for _ in range(self.h)]


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def draw_layer(self, y: int | str, x: int | str, layer: "Layer") -> None:
        """Draw another layer onto this layer at the specified coordinates."""
        h = min(layer.h, self.h)
        w = min(layer.w, self.w)
        layer = layer.copy()
        layer.set_size(h, w)
        y, x = self._parse_coords(h, w, y, x)
        self._stamp(y, x, layer._data, layer.blend_mode)

    # --------------------------------------------------------------------------
    def draw_matrix(self, y: int | str, x: int | str, chars: list[str], attrs: list[list[int]], blend = pr.BlendMode.OVERLAY) -> None:
        h = len(chars)
        w = len(chars[0])
        data = self.get_pixel_mat(chars, attrs)
        y, x = self._parse_coords(h, w, y, x)
        self._stamp(y, x, data, blend)

    # --------------------------------------------------------------------------
    def draw_text(self,
        y: int | str,  # Can be an integer or a string indicating position (e.g., 'T', 'C', 'B')
        x: int | str,  # Can be an integer or a string indicating position (e.g., 'L', 'C', 'R')
        string, # Accepts any object that can be converted to a string
        attr: int = None,
        blend = pr.BlendMode.OVERLAY,
        cut: dict[str, str] = {} # Cut dictionary to specify which edges to cut (e.g., {'T': 1, 'B': 2})
    ) -> None:
        """Draw a string at the specified coordinates with optional attributes and blending mode."""
        if attr is None: attr = pr.BLANK_ATTR

        rows = str(string).split('\n')
        h = min(len(rows), self.h)
        w = min(max(map(len, rows)), self.w)

        y, x = self._parse_coords(h, w, y, x)
        if (x >= self.w) or (y >= self.h): return

        chars = [row.ljust(w, pr.BLANK_CHAR)[:w] for row in rows[:h]]
        chars = self._parse_cut(y, x, cut, chars)
        attrs = [[attr for _ in row] for row in chars]
        data = self.get_pixel_mat(chars, attrs)
        self._stamp(y, x, data, blend)

    # --------------------------------------------------------------------------
    def draw_border(self,
        ls = '│', rs = '│', ts = '─', bs = '─',
        tl = '┌', tr = '┐', bl = '└', br = '┘',
        attr = None, blend = pr.BlendMode.OVERLAY
    ) -> None:
        """Draw a border around the layer with specified characters and attribute."""
        if attr is None: attr = pr.BLANK_ATTR

        h = self.h - 2
        w = self.w - 2
        BC = pr.BLANK_CHAR
        BA = pr.BLANK_ATTR
        data = self.get_pixel_mat(
            chars = \
                [tl + ts*w + tr]   +\
                [ls + BC*w + rs]*h +\
                [bl + bs*w + br],
            attrs = \
                [[attr] + [attr]*w + [attr]]   +\
                [[attr] + [ BA ]*w + [attr]]*h +\
                [[attr] + [attr]*w + [attr]]
        )
        self._stamp(0, 0, data, blend)


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def yield_render_data(self) -> Generator[tuple[str, int], None, None]:
        """Yield characters and their attributes ready for rendering."""
        flat_chars = ''.join(''.join(pixel._char for pixel in row) for row in self._data)
        flat_attrs = [pixel._attr for row in self._data for pixel in row]

        attrs_offset_0 = flat_attrs[:-1]
        attrs_offset_1 = flat_attrs[1:]

        attrs_mask = (a != b for a,b in zip(attrs_offset_0, attrs_offset_1))
        border_idxs = [0] + [i for i,a in enumerate(attrs_mask, start = 1) if a] + [len(flat_chars)]

        for i0,i1 in zip(border_idxs[:-1], border_idxs[1:]):
            yield flat_chars[i0:i1], flat_attrs[i0]


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def _parse_coords(self, h: int, w: int, y: int | str, x : int | str) -> tuple[int, int]:
        """Parse the coordinates for drawing, allowing for string-based positioning."""
        if isinstance(y, str):
            match y[0].upper():
                case 'T': yval = 0
                case 'C': yval = (self.h - h) // 2
                case 'B': yval = self.h - h
                case  _ : raise ValueError(f"Invalid y value: '{y}'")
            modifier = y[1:]
            if modifier: yval += int(modifier)
        else: yval = y

        if isinstance(x, str):
            match x[0].upper():
                case 'L': xval = 0
                case 'C': xval = (self.w - w) // 2
                case 'R': xval = self.w - w
                case  _ : raise ValueError(f"Invalid x value: '{y}'")
            modifier = x[1:]
            if modifier: xval += int(modifier)
        else: xval = x

        return yval, xval


    # --------------------------------------------------------------------------
    def _parse_cut(self, y: int, x: int, cut: dict[str, str], chars: list[str]) -> list[str]:
        """Parse a cut dictionary to adjust the characters matrix."""
        for k,v in cut.items():
            match k.upper():
                case 'T': chars = chars[v:]
                case 'B': chars = chars[:self.h-y-v]
                case 'L': chars = tuple(map(lambda row: row[v:], chars))
                case 'R': chars = tuple(map(lambda row: row[:self.w-x-v], chars))
                case  _ : raise ValueError(f"Invalid cut key: '{k}'")
        return chars


    # --------------------------------------------------------------------------
    def _stamp(self,
        y: int, x: int,
        data: list[list["pr.Pixel"]],
        blend = pr.BlendMode.OVERLAY
    ) -> None:
        """Stamp a matrix of pixels onto the layer at the specified coordinates."""
        if (y >= self.h) or (x >= self.w): return
        if not len(data): return

        match blend:
            case pr.BlendMode.OVERLAY:    func = pr.Pixel.overlay
            case pr.BlendMode.OVERWRITE:  func = pr.Pixel.overwrite
            case pr.BlendMode.MERGE_ATTR: func = pr.Pixel.merge_attr
            case _ : raise ValueError(f"Unknown blend_mode: {blend}")

        h = len(data)
        w = len(data[0])

        y0 = y; x0 = x
        y1 = min(y + h, self.h)
        x1 = min(x + w, self.w)

        mat_orig = self._data[y0:y1]
        mat_modf = data[:y1-y]

        self._data[y0:y1] = [
            row_orig[:x0] + [
                func(o,m) for o,m in zip(row_orig[x0:x1], row_modf[:x1-x])
            ] + row_orig[x1:]
            for row_orig, row_modf in zip(mat_orig, mat_modf)
        ]


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def _create_row(self, length: int) -> list["pr.Pixel"]:
        """Create a row of pixels with the specified length, initialized to blank."""
        return [pr.Pixel() for _ in range(length)]

    # --------------------------------------------------------------------------
    def _add_rows(self, n: int) -> None:
        """Add 'n' rows to the layer, initializing them with blank pixels."""
        self._data += [self._create_row(self.w) for _ in range(n)]

    # --------------------------------------------------------------------------
    def _add_cols(self, n: int) -> None:
        """Add 'n' columns to each row in the layer, initializing them with blank pixels."""
        self._data = [row + self._create_row(n) for row in self._data]

    # --------------------------------------------------------------------------
    def _remove_rows(self, n: int) -> None:
        """Remove rows after the first 'n' rows."""
        self._data = self._data[:n]

    # --------------------------------------------------------------------------
    def _remove_cols(self, n: int) -> None:
        """Remove columns after the first 'n' columns."""
        self._data = [row[:n] for row in self._data]


# //////////////////////////////////////////////////////////////////////////////



import prismatui as pr

# --------------------------------------------------------------------------
def save_layer(path_pri: str, layer: "pr.Layer") -> None:
    """Save a layer to a PRI file"""
    chars: list[str] = layer.get_chars_row_as_strs()
    pairs: list[list[int]] = layer.get_attrs()

    h = len(chars)
    w = len(chars[0]) if h > 0 else 0

    assert all(len(row) == w for row in chars), "All rows in 'chars' must have the same width."
    assert all(len(row) == w for row in pairs), "All rows in 'pairs' must have the same width as 'chars'."
    assert len(pairs) == h , "'pairs' must have the same height as 'chars'."

    with open(path_pri, "wb") as file:
        file.write(h.to_bytes(2, byteorder="little"))
        file.write(w.to_bytes(2, byteorder="little"))
        file.write(b'\n')
        file.write('\n'.join(chars).encode("utf-8"))
        file.write(b'\n')
        for row in pairs: file.write(bytes(int(x) for x in row))


# --------------------------------------------------------------------------
def load_layer(path_pri: str) -> "pr.Layer":
    """Load a layer from a PRI file."""
    with open(path_pri, "rb") as file:
        h = int.from_bytes(file.read(2), byteorder="little")
        w = int.from_bytes(file.read(2), byteorder="little")
        nchars = h * (w + 1) - 1 # +1 for breaklines (except the last one)

        file.read(1) # skip a breakline character
        chars = file.read(nchars).decode("utf-8")
        file.read(1) # skip a breakline character

        pairs = [[int(file.read(1)[0]) for _ in range(w)] for _ in range(h)]
        attrs = [[pr._CURRENT_BACKEND.get_color_pair(i) for i in row] for row in pairs]
    return pr.Layer(h, w, chars.split('\n'), attrs)


# --------------------------------------------------------------------------



from typing import Generator

import prismatui as pr

# //////////////////////////////////////////////////////////////////////////////
class Section:
    """A Section is a container for layers and child sections.
    It can be used to create a hierarchical structure of sections, each with its own layers.
    Each section has a size and position relative to its parent section
    (or the terminal, in case of the root section)."""
    def __init__(self):
        self._parent: "Section" = None

        self.h: int; self.w: int
        self.y: int; self.x: int
        self.hrel: int|float = 1.0
        self.wrel: int|float = 1.0
        self.yrel: int|float = 0
        self.xrel: int|float = 0
        self._update_dimensions()

        self._children: list["Section"] = []
        self._layers = [pr.Layer(self.h, self.w)]


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def set_parent(self, parent: "Section") -> None:
        """Set the parent section of this section."""
        self._parent = parent
        parent._children.append(self)
        self._update_dimensions()

    # --------------------------------------------------------------------------
    def create_child(self,
        hrel: int|float, wrel: int|float,
        yrel: int|float, xrel: int|float
    ) -> "Section":
        """Create a child section with the specified relative dimensions and position."""
        child = Section()
        child.hrel = hrel
        child.wrel = wrel
        child.yrel = yrel
        child.xrel = xrel
        child.set_parent(self)
        return child

    # --------------------------------------------------------------------------
    def create_layer(self) -> "pr.Layer":
        """Create a new layer in this section."""
        layer = pr.Layer(self.h, self.w)
        self._layers.append(layer)
        return layer

    # --------------------------------------------------------------------------
    def create_mosaic(self, layout: str, divider = '\n') -> dict:
        """Create a mosaic of sections based on the provided 'mosaic' layout
        string, where each character represents a section. Takes inspiration
        from matplotlib's `subplot_mosaic` function."""
        section_dict = {}
        for char, hwyx in pr.mosaic_parser(layout, divider).items():
            section = self.create_child(*hwyx)
            section_dict[char] = section
        return section_dict


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def get_size(self) -> tuple[int, int]:
        """Get the size of this section."""
        return self.h, self.w

    # --------------------------------------------------------------------------
    def get_position(self) -> tuple[int, int]:
        """Get the position of this section relative to its parent."""
        return self.y, self.x

    # --------------------------------------------------------------------------
    def get_bottom_layer(self) -> "pr.Layer":
        """Get the bottom layer of this section."""
        return self._layers[0]

    # --------------------------------------------------------------------------
    def get_top_layer(self) -> "pr.Layer":
        """Get the top layer of this section."""
        return self._layers[-1]


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def iter_children(self) -> Generator["Section", None, None]:
        """Iterate over the child sections of this section."""
        return iter(self._children)

    # --------------------------------------------------------------------------
    def iter_layers(self) -> Generator["pr.Layer", None, None]:
        """Iterate over the layers of this section."""
        return iter(self._layers)

    # --------------------------------------------------------------------------
    def clear(self) -> None:
        """Clear all layers and child sections in this section."""
        for layer in self.iter_layers():
            layer.clear()

        for child in self.iter_children():
            child.clear()

    # --------------------------------------------------------------------------
    def aggregate_layers(self, agg_layer: "pr.Layer") -> None:
        """Aggregate all layers and child sections into the provided aggregate layer."""
        for layer in self.iter_layers():
            agg_layer.draw_layer(self.y, self.x, layer)

        for child in self.iter_children():
            child.aggregate_layers(agg_layer)

    # --------------------------------------------------------------------------
    def update_size(self) -> None:
        """Update the size of this section and its layers, as well as of children sections, based on the relative dimensions."""
        self._update_dimensions()

        for layer in self.iter_layers():
            layer.set_size(self.h, self.w)

        for child in self.iter_children():
            child.update_size()


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def draw_layer(self, y: int | str, x: int | str, layer: "pr.Layer") -> None:
        """Draw another layer onto the top layer of this section, at the specified coordinates."""
        self.get_top_layer().draw_layer(y, x, layer)


    # --------------------------------------------------------------------------
    def draw_matrix(self, y: int | str, x: int | str, chars: list[str], attrs: list[list[int]], blend = pr.BlendMode.OVERLAY) -> None:
        """Draw a character matrix on the top layer of this section, at the specified coordinates with optional blending mode."""
        self.get_top_layer().draw_matrix(y, x, chars, attrs, blend)

    # --------------------------------------------------------------------------
    def draw_text(self,
        y: int | str, x: int | str, string,
        attr: int = None, blend = pr.BlendMode.OVERLAY,
        cut: dict[str, str] = {}
    ) -> None:
        """Draw a string on the top layer of this section, at the specified coordinates with optional attributes and blending mode."""
        self.get_top_layer().draw_text(y, x, string, attr, blend, cut)

    # --------------------------------------------------------------------------
    def draw_border(self,
        ls = '│', rs = '│', ts = '─', bs = '─',
        tl = '┌', tr = '┐', bl = '└', br = '┘',
        attr = None, blend = pr.BlendMode.OVERLAY
    ) -> None:
        """Draw a border around the top layer of this section with specified characters and attribute."""
        self.get_top_layer().draw_border(ls, rs, ts, bs, tl, tr, bl, br, attr, blend)


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def _update_dimensions(self) -> None:
        """Update the dimensions of this section based on its relative size and position."""
        if self._parent is None: # root section
            self.h, self.w = pr._CURRENT_BACKEND.get_size()
            self.y, self.x = 0, 0
            return

        h = self.hrel; w = self.wrel
        y = self.yrel; x = self.xrel

        if isinstance(h, float):
            self.h = round(h * self._parent.h)
        elif isinstance(h, int):
            if h < 0: h += self._parent.h
            self.h = min(h, self._parent.h)

        if isinstance(w, float):
            self.w = round(w * self._parent.w)
        elif isinstance(w, int):
            if w < 0: w += self._parent.w
            self.w = min(w, self._parent.w)

        if isinstance(y, float):
            self.y = self._parent.y + round(y * self._parent.h)
        elif isinstance(y, int):
            if y < 0: y += self._parent.h
            self.y = y + self._parent.y

        if isinstance(x, float):
            self.x = self._parent.x + round(x * self._parent.w)
        elif isinstance(x, int):
            if x < 0: x += self._parent.w
            self.x = x + self._parent.x

        y_outbounds = (self.y + self.h) - (self._parent.y + self._parent.h)
        x_outbounds = (self.x + self.w) - (self._parent.x + self._parent.w)

        if y_outbounds > 0: self.y -= y_outbounds
        if x_outbounds > 0: self.x -= x_outbounds


# //////////////////////////////////////////////////////////////////////////////



### Prisma follows the same constants as the curses module.
ERR = -1
OK = 0
A_ATTRIBUTES = 4294967040
A_NORMAL = 0
A_STANDOUT = 65536
A_UNDERLINE = 131072
A_REVERSE = 262144
A_BLINK = 524288
A_DIM = 1048576
A_BOLD = 2097152
A_ALTCHARSET = 4194304
A_INVIS = 8388608
A_PROTECT = 16777216
A_CHARTEXT = 255
A_COLOR = 65280
A_HORIZONTAL = 33554432
A_LEFT = 67108864
A_LOW = 134217728
A_RIGHT = 268435456
A_TOP = 536870912
A_VERTICAL = 1073741824
A_ITALIC = 2147483648
COLOR_BLACK = 0
COLOR_RED = 1
COLOR_GREEN = 2
COLOR_YELLOW = 3
COLOR_BLUE = 4
COLOR_MAGENTA = 5
COLOR_CYAN = 6
COLOR_WHITE = 7
BUTTON1_PRESSED = 2
BUTTON1_RELEASED = 1
BUTTON1_CLICKED = 4
BUTTON1_DOUBLE_CLICKED = 8
BUTTON1_TRIPLE_CLICKED = 16
BUTTON2_PRESSED = 64
BUTTON2_RELEASED = 32
BUTTON2_CLICKED = 128
BUTTON2_DOUBLE_CLICKED = 256
BUTTON2_TRIPLE_CLICKED = 512
BUTTON3_PRESSED = 2048
BUTTON3_RELEASED = 1024
BUTTON3_CLICKED = 4096
BUTTON3_DOUBLE_CLICKED = 8192
BUTTON3_TRIPLE_CLICKED = 16384
BUTTON4_PRESSED = 65536
BUTTON4_RELEASED = 32768
BUTTON4_CLICKED = 131072
BUTTON4_DOUBLE_CLICKED = 262144
BUTTON4_TRIPLE_CLICKED = 524288
BUTTON5_PRESSED = 2097152
BUTTON5_RELEASED = 1048576
BUTTON5_CLICKED = 4194304
BUTTON5_DOUBLE_CLICKED = 8388608
BUTTON5_TRIPLE_CLICKED = 16777216
BUTTON_SHIFT = 67108864
BUTTON_CTRL = 33554432
BUTTON_ALT = 134217728
ALL_MOUSE_EVENTS = 268435455
REPORT_MOUSE_POSITION = 268435456
KEY_BREAK = 257
KEY_DOWN = 258
KEY_UP = 259
KEY_LEFT = 260
KEY_RIGHT = 261
KEY_HOME = 262
KEY_BACKSPACE = 263
KEY_F0 = 264
KEY_F1 = 265
KEY_F2 = 266
KEY_F3 = 267
KEY_F4 = 268
KEY_F5 = 269
KEY_F6 = 270
KEY_F7 = 271
KEY_F8 = 272
KEY_F9 = 273
KEY_F10 = 274
KEY_F11 = 275
KEY_F12 = 276
KEY_F13 = 277
KEY_F14 = 278
KEY_F15 = 279
KEY_F16 = 280
KEY_F17 = 281
KEY_F18 = 282
KEY_F19 = 283
KEY_F20 = 284
KEY_F21 = 285
KEY_F22 = 286
KEY_F23 = 287
KEY_F24 = 288
KEY_F25 = 289
KEY_F26 = 290
KEY_F27 = 291
KEY_F28 = 292
KEY_F29 = 293
KEY_F30 = 294
KEY_F31 = 295
KEY_F32 = 296
KEY_F33 = 297
KEY_F34 = 298
KEY_F35 = 299
KEY_F36 = 300
KEY_F37 = 301
KEY_F38 = 302
KEY_F39 = 303
KEY_F40 = 304
KEY_F41 = 305
KEY_F42 = 306
KEY_F43 = 307
KEY_F44 = 308
KEY_F45 = 309
KEY_F46 = 310
KEY_F47 = 311
KEY_F48 = 312
KEY_F49 = 313
KEY_F50 = 314
KEY_F51 = 315
KEY_F52 = 316
KEY_F53 = 317
KEY_F54 = 318
KEY_F55 = 319
KEY_F56 = 320
KEY_F57 = 321
KEY_F58 = 322
KEY_F59 = 323
KEY_F60 = 324
KEY_F61 = 325
KEY_F62 = 326
KEY_F63 = 327
KEY_DL = 328
KEY_IL = 329
KEY_DC = 330
KEY_IC = 331
KEY_EIC = 332
KEY_CLEAR = 333
KEY_EOS = 334
KEY_EOL = 335
KEY_SF = 336
KEY_SR = 337
KEY_NPAGE = 338
KEY_PPAGE = 339
KEY_STAB = 340
KEY_CTAB = 341
KEY_CATAB = 342
KEY_ENTER = 343
KEY_SRESET = 344
KEY_RESET = 345
KEY_PRINT = 346
KEY_LL = 347
KEY_A1 = 348
KEY_A3 = 349
KEY_B2 = 350
KEY_C1 = 351
KEY_C3 = 352
KEY_BTAB = 353
KEY_BEG = 354
KEY_CANCEL = 355
KEY_CLOSE = 356
KEY_COMMAND = 357
KEY_COPY = 358
KEY_CREATE = 359
KEY_END = 360
KEY_EXIT = 361
KEY_FIND = 362
KEY_HELP = 363
KEY_MARK = 364
KEY_MESSAGE = 365
KEY_MOVE = 366
KEY_NEXT = 367
KEY_OPEN = 368
KEY_OPTIONS = 369
KEY_PREVIOUS = 370
KEY_REDO = 371
KEY_REFERENCE = 372
KEY_REFRESH = 373
KEY_REPLACE = 374
KEY_RESTART = 375
KEY_RESUME = 376
KEY_SAVE = 377
KEY_SBEG = 378
KEY_SCANCEL = 379
KEY_SCOMMAND = 380
KEY_SCOPY = 381
KEY_SCREATE = 382
KEY_SDC = 383
KEY_SDL = 384
KEY_SELECT = 385
KEY_SEND = 386
KEY_SEOL = 387
KEY_SEXIT = 388
KEY_SFIND = 389
KEY_SHELP = 390
KEY_SHOME = 391
KEY_SIC = 392
KEY_SLEFT = 393
KEY_SMESSAGE = 394
KEY_SMOVE = 395
KEY_SNEXT = 396
KEY_SOPTIONS = 397
KEY_SPREVIOUS = 398
KEY_SPRINT = 399
KEY_SREDO = 400
KEY_SREPLACE = 401
KEY_SRIGHT = 402
KEY_SRSUME = 403
KEY_SSAVE = 404
KEY_SSUSPEND = 405
KEY_SUNDO = 406
KEY_SUSPEND = 407
KEY_UNDO = 408
KEY_MOUSE = 409
KEY_RESIZE = 410
KEY_MIN = 257
KEY_MAX = 511

### Additionally, ASCII characters (32-126) are provided.
KEY_SPACE  = 32
KEY_EXCLAMATION = 33
KEY_QUOTATION = 34
KEY_NUMBER = 35
KEY_DOLLAR = 36
KEY_PERCENT = 37
KEY_AMPERSAND = 38
KEY_APOSTROPHE = 39
KEY_L_PARENTHESIS = 40
KEY_R_PARENTHESIS = 41
KEY_ATERISK = 42
KEY_PLUS = 43
KEY_COMMA = 44
KEY_HYPHEN = 45
KEY_PERIOD = 46
KEY_SLASH = 47
KEY_0 = 48
KEY_1 = 49
KEY_2 = 50
KEY_3 = 51
KEY_4 = 52
KEY_5 = 53
KEY_6 = 54
KEY_7 = 55
KEY_8 = 56
KEY_9 = 57
KEY_COLON = 58
KEY_SEMICOLON = 59
KEY_LESS_THAN = 60
KEY_EQUAL = 61
KEY_GREATER_THAN = 62
KEY_QUESTION = 63
KEY_AT = 64
KEY_A_UPPER = 65
KEY_B_UPPER = 66
KEY_C_UPPER = 67
KEY_D_UPPER = 68
KEY_E_UPPER = 69
KEY_F_UPPER = 70
KEY_G_UPPER = 71
KEY_H_UPPER = 72
KEY_I_UPPER = 73
KEY_J_UPPER = 74
KEY_K_UPPER = 75
KEY_L_UPPER = 76
KEY_M_UPPER = 77
KEY_N_UPPER = 78
KEY_O_UPPER = 79
KEY_P_UPPER = 80
KEY_Q_UPPER = 81
KEY_R_UPPER = 82
KEY_S_UPPER = 83
KEY_T_UPPER = 84
KEY_U_UPPER = 85
KEY_V_UPPER = 86
KEY_W_UPPER = 87
KEY_X_UPPER = 88
KEY_Y_UPPER = 89
KEY_Z_UPPER = 90
KEY_L_SQUARE_BRACKET = 91
KEY_BACKSLASH = 92
KEY_R_SQUARE_BRACKET = 93
KEY_CARET = 94
KEY_UNDERSCORE = 95
KEY_GRAVE = 96
KEY_A_LOWER = 97
KEY_B_LOWER = 98
KEY_C_LOWER = 99
KEY_D_LOWER = 100
KEY_E_LOWER = 101
KEY_F_LOWER = 102
KEY_G_LOWER = 103
KEY_H_LOWER = 104
KEY_I_LOWER = 105
KEY_J_LOWER = 106
KEY_K_LOWER = 107
KEY_L_LOWER = 108
KEY_M_LOWER = 109
KEY_N_LOWER = 110
KEY_O_LOWER = 111
KEY_P_LOWER = 112
KEY_Q_LOWER = 113
KEY_R_LOWER = 114
KEY_S_LOWER = 115
KEY_T_LOWER = 116
KEY_U_LOWER = 117
KEY_V_LOWER = 118
KEY_W_LOWER = 119
KEY_X_LOWER = 120
KEY_Y_LOWER = 121
KEY_Z_LOWER = 122
KEY_L_CURLY_BRACE = 123
KEY_VERTICAL_BAR = 124
KEY_L_CURLY_BRACE = 125
KEY_TILDE = 126



# //////////////////////////////////////////////////////////////////////////////
class DebugLogger:
    """Debug logger class to log messages to a file.
    It creates a log file at the specified path and appends messages to it.
    Example:
        from prisma.utils import Debug;
        d = Debug("logs/name.log")
        d.log("Some value:", value)
    """
    def __init__(self, path: str):
        import datetime
        self.path = path
        with open(self.path, 'w') as file:
            file.write(f"{datetime.datetime.now()}\n\n")

    # --------------------------------------------------------------------------
    def log(self, *values, sep = ' ', end = '\n'):
        """Log values to the debug file.
        Args:
            *values: Values to log.
            sep (str): Separator between values. Default is a space.
            end (str): String appended after the last value. Default is a newline.
        """
        text = sep.join(map(str, values))
        with open(self.path, 'a') as file:
            file.write(text + end)


# //////////////////////////////////////////////////////////////////////////////



from collections import OrderedDict

# ------------------------------------------------------------------------------
def mosaic_parser(layout: str, divider = '\n') -> dict:
    """Parse a mosaic layout string and return a dictionary with the dimensions
    and positions of each character in the layout."""
    if not layout: return {}

    rows = layout.split(divider)
    cols = tuple(zip(*rows))

    row_lenghts = map(len, rows)
    if len(set(row_lenghts)) != 1:
        raise ValueError("Not all mosaic rows have the same lenght.")

    row_idxs = tuple(range(len(row)) for row in rows)
    col_idxs = tuple(range(len(col)) for col in cols)

    chars = set(layout)
    if divider in chars: chars.remove(divider)

    h_mosaic = len(rows)
    w_mosaic = len(cols)

    data = OrderedDict()
    for char in sorted(chars):
        masked_row_idxs = _apply_mask(row_idxs, rows, char)
        masked_col_idxs = _apply_mask(col_idxs, cols, char)

        err_prefix = "Error parsing mosaic layout:"
        if not _all_elements_equal(masked_row_idxs):
            raise ValueError(f"{err_prefix} Not all rows for char '{char}' have the same length.")

        if not _all_elements_equal(masked_col_idxs):
            raise ValueError(f"{err_prefix} Not all columns for char '{char}' have the same length.")

        if not _is_sequential(masked_row_idxs[0]):
            raise ValueError(f"{err_prefix} Rows of char '{char}' are interrupted.")

        if not _is_sequential(masked_col_idxs[0]):
            raise ValueError(f"{err_prefix} Columns of char '{char}' are interrupted.")

        y_char = masked_col_idxs[0][0] / h_mosaic
        x_char = masked_row_idxs[0][0] / w_mosaic
        h_char = len(masked_col_idxs[0]) / h_mosaic
        w_char = len(masked_row_idxs[0]) / w_mosaic
        data[char] = (h_char, w_char, y_char, x_char)

    return data

# ------------------------------------------------------------------------------
def _apply_mask(idxs: list[list[int]], mat: list[list[str]], char: str) -> tuple[tuple[int]]:
    """Apply a mask to the matrix to filter out elements that match the given character."""
    mat_out = tuple(map(
        lambda idx, arr: tuple(filter(
            lambda tup: tup[1] == char,
            zip(idx, arr)
        )),
        idxs, mat
    ))
    return tuple(tuple(n[0] for n in arr) for arr in mat_out if arr)

# ------------------------------------------------------------------------------
def _all_elements_equal(iterable: iter) -> bool:
    """Check if all elements in the iterable are equal."""
    iterator = iter(iterable)
    try: first = next(iterator)
    except StopIteration: return True
    return all(first == rest for rest in iterator)

# ------------------------------------------------------------------------------
def _is_sequential(iterable: iter) -> bool:
    """Check if the elements in the iterable are sequential, starting from the first element."""
    iterator = iter(iterable)
    try: first = next(iterator)
    except StopIteration: return True
    return all(i == element for i,element in enumerate(iterator, start = first + 1))

# ------------------------------------------------------------------------------

################################################################################
if __name__ == "__main__":
    #### test the parser

    layouts = dict(
        empty = '',
        row_clean = '\n'.join((
            "aaab",
        )),
        row_bad = '\n'.join((
            "axab",
        )),
        col_clean = '\n'.join((
            "a",
            "a",
            "c",
        )),
        col_bad = '\n'.join((
            "a",
            "x",
            "a",
            "c",
        )),
        box_clean = '\n'.join((
            "ltttt",
            "laaab",
            "laaab",
            "lcddd",
        )),
        box_missing_char = '\n'.join((
            "ltttt",
            "laaab",
            "laaa",
            "lcddd",
        )),
        box_bad_corner = '\n'.join((
            "ltttt",
            "laaab",
            "laaxb",
            "lcddd",
        )),
        box_bad_row = '\n'.join((
            "ltttt",
            "laaab",
            "lxxxb",
            "laaab",
            "lcddd",
        )),
        box_bad_col = '\n'.join((
            "ltttt",
            "laxab",
            "laxab",
            "lcddd",
        )),
        box_bad_interior = '\n'.join((
            "ltttt",
            "laaab",
            "laxab",
            "laaab",
            "lcddd",
        )),
    )

    for k, v in layouts.items():
        print("------------------", k)
        print(v); print()
        try:
            hwyx_data = mosaic_parser(v)
        except ValueError as e:
            print("XXX", e)
        else:
            print(*sorted(hwyx_data.items()), sep = '\n')
        print()


################################################################################



from ._backend.backend import Backend
from ._backend.backend_curses import BackendCurses
from ._backend.constants import *

# --------------------------------------------------------------------------
_CURRENT_BACKEND: "Backend | None" = None
BLANK_CHAR = ' '
BLANK_ATTR = A_NORMAL

MAX_PALETTE_COLORS = 256
MAX_PALETTE_PAIRS  = 256
ALPHA_THRESHOLD = 128

# --------------------------------------------------------------------------
from ._misc.blend_mode import BlendMode
from ._misc.palette import Palette
from ._misc.parser_pri import load_layer, save_layer

from ._utils.debug_logger import DebugLogger
from ._utils.mosaic_parser import mosaic_parser

from ._components.pixel import Pixel
from ._components.layer import Layer
from ._components.section import Section
from ._components.terminal import Terminal

# --------------------------------------------------------------------------
def set_backend(backend: str|Backend) -> None:
    global _CURRENT_BACKEND

    if isinstance(backend, Backend):
        _CURRENT_BACKEND = backend

    elif isinstance(backend, str):
        match backend:
            case "curses": _CURRENT_BACKEND = BackendCurses()
            case _: raise ValueError(f"Unknown backend: {backend}")


# --------------------------------------------------------------------------
def set_nodelay(boolean: bool) -> None:
    """Set the nodelay mode for the backend. When enabled, get_key() will not block the terminal."""
    _CURRENT_BACKEND.set_nodelay(boolean)

def sleep(ms: int) -> None:
    """Sleep for a given number of milliseconds."""
    _CURRENT_BACKEND.sleep(ms)

def write_text(y: int, x: int, chars: str, attr: int = BLANK_ATTR) -> None:
    """Write text to the terminal at a specific position."""
    _CURRENT_BACKEND.write_text(y, x, chars, attr)

def get_size(update: bool = False) -> tuple[int, int]:
    """Get the size of the terminal, as (COLS, LINES).
    If update is True, a backend method will be called to update the current COLS and LINES values."""
    return _CURRENT_BACKEND.get_size(update)

def supports_color() -> bool:
    """Check if the backend supports color."""
    return _CURRENT_BACKEND.supports_color()

def init_color(i: int, r: int, g: int, b: int) -> None:
    """Initialize a color with index 'i' and values 'r','g','b' to be used in color pairs by the terminal."""
    _CURRENT_BACKEND.init_color(i, r, g, b)

def get_color_pair(i: int) -> int:
    """Retrieve the color pair for a given index."""
    return _CURRENT_BACKEND.get_color_pair(i)

def init_pair(i: int, fg: int, bg: int) -> int:
    """Initialize a color pair (fg,bg) which can be accessed with index 'i'. Returns the color pair value."""
    _CURRENT_BACKEND.init_pair(i, fg, bg)
    return get_color_pair(i)


# --------------------------------------------------------------------------
set_backend("curses")



import prismatui as pr

# //////////////////////////////////////////////////////////////////////////////
class Pixel:
    """A Pixel objects is a container for a single character with an associated attribute."""
    def __init__(self, char = pr.BLANK_CHAR, attr = pr.BLANK_ATTR):
        self._char = char
        self._attr = attr


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def overwrite(self, other: "Pixel") -> "Pixel":
        """The other pixel replaces this pixel unconditionally."""
        return other

    # --------------------------------------------------------------------------
    def overlay(self, other: "Pixel") -> "Pixel":
        """If the other pixel is not blank, it replaces this pixel.
        Blank pixels are therefore "transparent" and don't affect the pixels under them."""
        return self if other._is_blank() else other

    # --------------------------------------------------------------------------
    def merge_attr(self, other: "Pixel") -> "Pixel":
        """The other pixel's attribute is merged with this pixel's attribute,
        by applying a bitwise OR operation."""
        self._attr |= other._attr
        return self


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def _is_blank(self) -> bool:
        """A pixel is considered to be "blank" when BOTH its character and attribute
        are set to their blank values."""
        return (self._char == pr.BLANK_CHAR) and (self._attr == pr.BLANK_ATTR)


# //////////////////////////////////////////////////////////////////////////////



import prismatui as pr

# //////////////////////////////////////////////////////////////////////////////
class Terminal:
    """The Terminal class is the main entry point for creating a TUI application.
    It manages the terminal size, input handling, rendering of sections and color palette.
    It handles application lifecycle events such as start, resize, update and end."""
    def __init__(self):
        self.h: int = 0
        self.w: int = 0
        self.key: int = -1
        self.root: pr.Section
        self.palette: pr.Palette

        self._no_delay: bool = False
        self._nap_ms: int = 0
        self._wait = lambda: None
        self._running: bool = False


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def run(self, fps: int = 0) -> None:
        """Main loop of the terminal application.
        The fps parameter controls the frame rate of the application.
        If fps is 0, the application runs in no-delay mode, meaning it will not wait between frames.
        If fps is greater than 0, the application will wait for the required number of milliseconds
        between frames to reach the indicated frame rate."""
        self.set_fps(fps)
        try:
            pr._CURRENT_BACKEND._start()
            self._on_start()
            while self._running:
                self._on_resize()
                self._on_update()
            self._on_end()
        finally:
            pr._CURRENT_BACKEND._end()

    # --------------------------------------------------------------------------
    def stop(self) -> None:
        """Stop the terminal application."""
        self._running = False

    # --------------------------------------------------------------------------
    def fetch_key(self) -> int:
        """Fetch the next key from the terminal input.
        This method will block until a key is pressed, unless the terminal is in no-delay mode."""
        self.key = pr._CURRENT_BACKEND._get_key()
        return self.key

    # --------------------------------------------------------------------------
    def exhaust_keys(self) -> None:
        """Exhaust all keys from the terminal input buffer.
        Attention: calling to fetch_key()->prisma._CURRENT_BACKEND._get_key() internally rellies on stdscr.getch(),
        so this method will block the terminal when fps=0 (i.e. outside no-delay mode)."""
        while self.fetch_key() != -1:
            pass


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def set_fps(self, fps: int) -> None:
        """Set the frames per second (fps) for the terminal application.
        If fps is 0, the application will run in no-delay mode, meaning it will not wait between frames.
        If fps is greater than 0, the application will wait for the required number of milliseconds
        between frames to reach the indicated frame rate."""
        if not fps:
            self._no_delay = False
            self._nap_ms = 0
            self._wait = lambda: None
        else:
            self._no_delay = True
            self._nap_ms = int(1000 / fps)
            self._wait = lambda: pr._CURRENT_BACKEND.sleep(self._nap_ms)

    # --------------------------------------------------------------------------
    def get_size(self) -> tuple[int, int]:
        """Get the current size of the terminal."""
        return self.h, self.w

    # --------------------------------------------------------------------------
    def resize_terminal(self, h: int, w: int) -> None:
        """Resize the terminal to the specified height and width. Might not work in some terminals."""
        print(f"\x1b[8;{h};{w}t")

    # --------------------------------------------------------------------------
    def draw_layer(self, y: int | str, x: int | str, layer: "pr.Layer") -> None:
        """Draw another layer onto the top layer of the root section, at the specified coordinates."""
        self.root.draw_layer(y, x, layer)

    # --------------------------------------------------------------------------
    def draw_matrix(self, y: int | str, x: int | str, chars: list[str], attrs: list[list[int]], blend = pr.BlendMode.OVERLAY) -> None:
        """Draw a character matrix on the top layer of the root section, at the specified coordinates with optional blending mode."""
        self.root.draw_matrix(y, x, chars, attrs, blend)

    # --------------------------------------------------------------------------
    def draw_text(self,
        y: int | str, x: int | str, string,
        attr: int = None, blend = pr.BlendMode.OVERLAY,
        cut: dict[str, str] = {}
    ) -> None:
        """Draw a string on the top layer of the root section, at the specified coordinates with optional attributes and blending mode."""
        self.root.draw_text(y, x, string, attr, blend, cut)

    # --------------------------------------------------------------------------
    def draw_border(self,
        ls = '│', rs = '│', ts = '─', bs = '─',
        tl = '┌', tr = '┐', bl = '└', br = '┘',
        attr = None, blend = pr.BlendMode.OVERLAY
    ) -> None:
        """Draw a border around the top layer of the root section with specified characters and attribute."""
        self.root.draw_border(ls, rs, ts, bs, tl, tr, bl, br, attr, blend)


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def on_start(self) -> None:
        """Called when the terminal application starts.
        This method can be overridden by the user to perform custom initialization."""
        return # overridden by user

    # --------------------------------------------------------------------------
    def on_resize(self) -> None:
        """Called when the terminal is resized.
        This method can be overridden by the user to perform custom actions on resize."""
        return # overridden by user

    # --------------------------------------------------------------------------
    def on_update(self) -> None:
        """Called on each frame update.
        This method can be overridden by the user to perform custom actions on each frame."""
        return # overridden by user

    # --------------------------------------------------------------------------
    def on_end(self) -> None:
        """Called when the terminal application ends.
        This method can be overridden by the user to perform custom cleanup actions."""
        return # overridden by user

    # --------------------------------------------------------------------------
    def should_stop(self) -> bool:
        """Determine whether the terminal application should stop.
        This method can be overridden by the user to implement custom stop conditions.
        By default, it returns False, meaning the application will continue running."""
        return False # overridden by user


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def _on_start(self) -> None:
        """Internal method called when the terminal application starts.
        It initializes the terminal size, root section, palette
        and sets the backend to no-delay mode or not, according to the provided fps."""
        self.root = pr.Section()
        self.palette = pr.Palette()
        pr._CURRENT_BACKEND.set_nodelay(self._no_delay)

        self._running = True
        self.on_start()

    # --------------------------------------------------------------------------
    def _on_resize(self) -> None:
        """Internal method called when the terminal is resized."""
        h,w = pr._CURRENT_BACKEND.get_size(update = True)

        if (self.h == h) and (self.w == w): return

        self.h = h; self.w = w
        self.root.update_size()
        pr._CURRENT_BACKEND._resize(self.h, self.w)
        self.on_resize()

    # --------------------------------------------------------------------------
    def _on_update(self) -> None:
        """Internal method called on each frame update.
        It clears the root section, calls the on_update method,
        and renders the current state of the terminal."""
        self.root.clear()
        self.on_update()
        self._render()

        self.key = pr._CURRENT_BACKEND._get_key()
        if self.should_stop(): self.stop()
        self._wait()

    # --------------------------------------------------------------------------
    def _on_end(self) -> None:
        """Internal method called when the terminal application ends."""
        self.on_end()


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def _render(self) -> None:
        """Render the current state of the terminal by aggregating all layers
        and writing the rendered data to the terminal backend."""
        master_layer = pr.Layer(self.h, self.w)
        self.root.aggregate_layers(master_layer)

        idx = 0
        for chars,attr in master_layer.yield_render_data():
            y,x = divmod(idx, self.w)
            pr._CURRENT_BACKEND.write_text(y, x, chars, attr)
            idx += len(chars)

        pr._CURRENT_BACKEND._refresh()


# //////////////////////////////////////////////////////////////////////////////



from enum import Enum, auto

# //////////////////////////////////////////////////////////////////////////////
class BlendMode(Enum):
    """Enum for layer blending modes. Defines how layers are blended when drawn."""
    OVERLAY = auto() # Default blending mode, overlays the layer on top of the existing pixels.
    OVERWRITE = auto() # Overwrites the existing pixels with the new layer's pixels.
    MERGE_ATTR = auto() # Merges the attributes of the new layer with the existing pixels, preserving the characters.
